library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

entity opb_xsb300 is
  generic (
    C_OPB_AWIDTH    : integer := 32;
    C_OPB_DWIDTH    : integer := 32;
    C_BASEADDR      : std_logic_vector := X"2000_0000";
    C_HIGHADDR      : std_logic_vector := X"2000_00FF"
    );

    Port ( 	OPB_Clk : in std_logic;
				OPB_Rst : in std_logic;
				OPB_ABus : in std_logic_vector (31 downto 0);
				OPB_BE : in std_logic_vector (3 downto 0);
				OPB_DBus : in std_logic_vector (31 downto 0);
				OPB_RNW : in std_logic;
				OPB_select : in std_logic;
				OPB_seqAddr : in std_logic;
				UIO_DBus : out std_logic_vector (31 downto 0);
				UIO_errAck : out std_logic;
				UIO_retry : out std_logic;
				UIO_toutSup : out std_logic;
				UIO_xferAck : out std_logic;
				PB_A : out std_logic_vector (19 downto 0);
				PB_UB_N : out std_logic; 
				PB_LB_N : out std_logic;
				PB_WE_N : out std_logic;
				PB_OE_N : out std_logic;
				RAM_CE_N : out std_logic;
				PB_D : inout std_logic_vector (15 downto 0);
                                CS0_N : out std_logic;
                                CS1_N : out std_logic;
                                IOCS16_N : in std_logic;
                                IORDY : in std_logic;
                                INTRQ : in std_logic
	 		   );
end opb_xsb300;

architecture Behavioral of opb_xsb300 is

constant C_MASK : integer := 0; -- huge address window as we are a bridge


signal addr_mux : std_logic_vector(19 downto 0);

signal i : integer;

signal onecycle : std_logic ;
signal amuxsel : std_logic := '1';
signal hihalf : std_logic;
signal rce0, rce1, rreset : std_logic;
signal xfer : std_logic;
signal pb_wr, pb_rd : std_logic;

signal sram_ce : std_logic;

signal rnw : std_logic;

signal addr : std_logic_vector (23 downto 0); 	

signal be : std_logic_vector (3 downto 0);
signal pb_bytesel : std_logic_vector (1 downto 0);

signal wdata : std_logic_vector (31 downto 0);
signal wdata_mux : std_logic_vector (15 downto 0);

signal rdata : std_logic_vector (15 downto 0);	-- register data read - FDRE

signal cs : std_logic;    -- sram cs
signal cf_cs : std_logic; -- cf cs

signal cf_busy : std_logic;
signal cf_rdy : std_logic;
signal cf_intrq : std_logic;

signal cf_out_data : std_logic_vector( 15 downto 0 );

signal cf_addr : std_logic_vector( 19 downto 0 );
signal sram_addr : std_logic_vector( 19 downto 0 );

component memoryctrl 
    Port ( rst : in std_logic;
           clk : in std_logic;
           cs : in std_logic;
           select0 : in std_logic;
           rnw : in std_logic;
           onecycle : in std_logic;
           hihalf : out std_logic;
           pb_wr : out std_logic;
           pb_rd : out std_logic;
           xfer : out std_logic;
           ce0 : out std_logic;
           ce1 : out std_logic;
           rres : out std_logic);
end component;

component pad_io 
    Port ( clk : in std_logic;
           rst : in std_logic;
           PB_A : out std_logic_vector(19 downto 0);
           PB_UB_N : out std_logic;
           PB_LB_N : out std_logic;
           PB_WE_N : out std_logic;
           PB_OE_N : out std_logic;
           RAM_CE_N : out std_logic;
           PB_D : inout std_logic_vector(15 downto 0);
           pb_addr : in std_logic_vector(19 downto 0);
           pb_ub : in std_logic;
           pb_lb : in std_logic;
           pb_wr : in std_logic;
           pb_rd : in std_logic;
           ram_ce : in std_logic;
           pb_dread : out std_logic_vector(15 downto 0);
           pb_dwrite : in std_logic_vector(15 downto 0);
           CS0_N : out std_logic;
           CS1_N : out std_logic;
           IOCS16_N : in std_logic;
           IORDY : in std_logic;
           INTRQ : in std_logic;
           cf_cs_pad0 : in std_logic;
           cf_cs_pad1 : in std_logic;
           cf_iordy : out std_logic
           );
end component;

component cf
  Port (
    OPB_Clk : in std_logic;
    OPB_Rst : in std_logic;
    opb_addr : in std_logic_vector( 31 downto 0 );
    cf_data : in std_logic_vector( 15 downto 0 );     -- data from pad_io (cf)
    cf_cs : in std_logic;
    cf_rdy : in std_logic;                            -- iordy signal from pad_io (cf)
    opb_data : inout std_logic_vector( 31 downto 0 );
    cf_addr : out std_logic_vector( 19 downto 0 );
    cf_ce0 : out std_logic;
    cf_ce1 : out std_logic;
    cf_oe : out std_logic;
    cf_busy : out std_logic;
    cf_xfer : out std_logic
);
end component;

begin

compactflash : cf
port map (
    OPB_Clk => OPB_Clk,
    OPB_Rst => OPB_Rst,
    opb_addr => OPB_Abus,
    cf_data => PB_D,
    cf_cs => cf_cs,
    cf_rdy => cf_iordy,
    opb_data => cf_out_data,
    cf_addr => cf_addr,
    cf_ce0 : out std_logic;
    cf_ce1 : out std_logic;
    cf_oe : out std_logic;
    cf_busy : out std_logic;
    cf_xfer : out std_logic
);

-- the controller state machine

memoryctrl1 : memoryctrl
port map	(	rst => OPB_Rst,
				clk => OPB_Clk,
				cs => cs, 
				select0 => OPB_select, 
				rnw => rnw, 
				onecycle => onecycle,
				hihalf => hihalf, 
				pb_wr => pb_wr, 
				pb_rd => pb_rd, 
				xfer => xfer,
				ce0 => rce0, 
				ce1 => rce1, 
				rres => rreset);

-- compact flash state machine
-- compact flash
-- PADS

pad_io1 : pad_io
port map (
	clk => OPB_Clk,
	rst => OPB_Rst,
	PB_A => PB_A,
	PB_UB_N => PB_UB_N,	
	PB_LB_N => PB_LB_N,
	PB_WE_N => PB_WE_N,
	PB_OE_N => PB_OE_N, 
	RAM_CE_N => RAM_CE_N,
	PB_D => PB_D,
	pb_addr => addr_mux,
	pb_rd => pb_rd,
	pb_wr => pb_wr,
	pb_ub => pb_bytesel(1),
	pb_lb => pb_bytesel(0),
	ram_ce => sram_ce,
	pb_dread  => rdata,
	pb_dwrite => wdata_mux,
	CS0_N => CS0_N,
	CS1_N => CS1_N,
	IOCS16_N => IOCS16_N,
	IORDY => IORDY,
	INTRQ => INTRQ,
	cf_cs_pad0 => cf_ce0,
	cf_cs_pad1 => cf_ce1,
	cf_rdy => cf_rdy
);

	sram_ce <= (pb_rd or pb_wr) and (not cf_busy);
	amuxsel <= '0';

	--addr_mux <= addr(20 downto 2) & (addr(1) or hihalf);
	sram_addr <= addr(20 downto 2) & (addr(1) or hihalf);

	onecycle <= (not be(3)) or (not be(2)) or (not be(1)) or (not be(0));

	wdata_mux <= wdata(15 downto 0) when ((addr(1) or hihalf) = '1') else wdata(31 downto 16);

process(be, addr(1), hihalf, pb_rd, pb_wr)
begin

if pb_rd='1' or pb_wr='1' then
    if addr(1)='1' or hihalf='1' then
      pb_bytesel <= be(1 downto 0);
    else
      pb_bytesel <= be(3 downto 2);
    end if;
  else
    pb_bytesel <= "00";
end if;

  
end process;
      

process( OPB_Clk, OPB_Rst, OPB_ABus, cf_addr, cf_busy )
begin
	if OPB_Clk'event and OPB_Clk ='1' then
		if cf_busy = '1' then
			addr_mux <= cf_addr;
		else
			addr_mux <= sram_addr;
		end if;
	end if;
end process;
                      

process (OPB_Clk,	OPB_Rst,	OPB_ABus, OPB_BE, OPB_DBus, 
			OPB_RNW, OPB_select, OPB_seqAddr)

begin

--	for i in 0 to 1 loop
--		if (addr(1) or hihalf) = '1' then
--			pb_bytesel(i) <= ( be(i) and (pb_rd or pb_wr)) or videocycle; 
--		else 
--			pb_bytesel(i) <= ( be(i + 2) and (pb_rd or pb_wr)) or videocycle;
--		end if;
--	end loop;

	if (OPB_ABus(31 downto 20) = X"008") then
		if cf_busy = '0' then
			cs <= OPB_select;
			cf_cs <= '0';
		end if;
	elsif (OPB_ABus(31 downto 24) = X"01") then
		cs <= '0';
		cf_cs <= OPB_select;
	else
		cs <= '0';
		cf_cs <= '0';
	end if;

-- register rw
	if OPB_Clk'event and OPB_Clk ='1' then
	
		if OPB_Rst = '1' then
			rnw <= '0'; 
		else 
			rnw <= OPB_RNW;
		end if;
	
	end if;

-- register adresses A23 .. A0
	if OPB_Clk'event and OPB_Clk ='1' then
	   for i in 0 to 23 loop
			if OPB_Rst = '1' then
  				addr(i) <= '0';
			else
				addr(i) <= OPB_ABus(i);
			end if;	
   	end loop;
	end if;
			
	-- register BE
	if OPB_Clk'event and OPB_Clk ='1' then
			if OPB_Rst = '1' then
  				be <= "0000";
			else
				be <= OPB_BE;
			end if;
	end if;

-- register data write
	if OPB_Clk'event and OPB_Clk ='1' then
	   for i in 0 to 31 loop
			if OPB_Rst = '1' then
  				wdata(i) <= '0';
			else
				wdata(i) <= OPB_DBus(i);
			end if;		
   	end loop;
	end if;

-- the fun begins
-- ce0/ce1 enables writing MSB (low) / LSB (high) halves

--always @(posedge OPB_Rst or posedge OPB_Clk) begin (synchrounous or asynchronous reset??) 

  		for i in 0 to 15 loop
			if OPB_Rst = '1' then
				UIO_DBus(i) <= '0';
			
			elsif OPB_Clk'event and OPB_Clk ='1' then 
				if rreset = '1' then
					UIO_DBus(i) <= '0';
				elsif (rce1 or rce0) = '1' then
					UIO_DBus(i) <= rdata(i);
				end if;
			end if;
		end loop;

		for i in 16 to 31 loop 
			if OPB_Rst = '1' then
				UIO_DBus(i) <= '0';
			elsif OPB_Clk'event and OPB_Clk ='1' then 

				if rreset = '1' then
					UIO_DBus(i) <= '0';
				elsif rce0 = '1' then
					UIO_DBus(i) <= rdata(i-16);
				end if;
			end if;
		end loop;

-- tie unused to ground
UIO_errAck <= '0';
UIO_retry <= '0';
UIO_toutSup <= '0';

UIO_xferAck <= xfer;

end process;

end Behavioral;
